////////////////////////////////////////////////////////////////////////////////
//  Colin McRae Rally 2005 Archive Decompressor                               //
//  Copyright (c) 2004  Game-Editing.net                                      //
//                                                                            //
//  This program is free software; you can redistribute it and/or modify      //
//  it under the terms of the GNU General Public License as published by      //
//  the Free Software Foundation; either version 2 of the License, or         //
//  (at your option) any later version.                                       //
//                                                                            //
//  This program is distributed in the hope that it will be useful,           //
//  but WITHOUT ANY WARRANTY; without even the implied warranty of            //
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             //
//  GNU General Public License for more details.                              //
//                                                                            //
//  You should have received a copy of the GNU General Public License         //
//  along with this program; if not, write to the Free Software               //
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA //
//                                                                            //
//  Written by Dan Strandberg & Ryan Foss                                     //
//                                                                            //
////////////////////////////////////////////////////////////////////////////////
unit FormMain;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, ShellApi, Menus;

type
  TfrmMain = class(TForm)
    OpenDialog: TOpenDialog;
    SaveDialog: TSaveDialog;
    edtSource: TEdit;
    edtDest: TEdit;
    btnBrowseForSource: TButton;
    btnBrowseForDest: TButton;
    btnDecompress: TButton;
    lblHomepage: TLabel;
    MainMenu: TMainMenu;
    Help1: TMenuItem;
    About1: TMenuItem;
    File1: TMenuItem;
    Exit1: TMenuItem;
    procedure btnDecompressClick(Sender: TObject);
    procedure btnBrowseForSourceClick(Sender: TObject);
    procedure btnBrowseForDestClick(Sender: TObject);
    procedure lblHomepageClick(Sender: TObject);
    procedure About1Click(Sender: TObject);
    procedure Exit1Click(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
  end;

const
  BUFFER_HIGH = $fff;
  BIGC_MARKER = 1128745282;

var
  frmMain: TfrmMain;

function Decompress(Source, Dest: TStream): Integer;

implementation

{$R *.dfm}

function Decompress(Source, Dest: TStream): Integer;
var
  x,
  BufferPos,
  FlagMask,
  Flag,
  UncompSize,
  CompSize,
  BytesWritten: Integer;
  Buffer: array [0..BUFFER_HIGH] of Byte;
  Len, Idx: Integer;
  b1,b2,c1: Byte;
begin
  // Read header
  Source.Read(x, 4);
  if x <> BIGC_MARKER then
    raise Exception.Create('Not a compressed/valid archive.');
  Source.Read(UncompSize, 4);
  Source.Read(CompSize, 4);

  BytesWritten := 0;
  BufferPos := 0;
  FlagMask := 1;
  Source.Read(Flag, 1);

  while BytesWritten < UncompSize do
  begin
    // If current bit in the Flag is 1 then the next byte is "uncompressed"
    // data that we just can read as is.
    if FlagMask and Flag <> 0 then
    begin
      Source.Read(c1, 1);
      //Write byte both to the buffer and the dest stream.
      Dest.Write(c1, 1);
      Buffer[BufferPos] := c1;
      BufferPos := (BufferPos + 1) and BUFFER_HIGH;
      Inc(BytesWritten);
    end else
    // Flag bit was 0 which means this is a word describing what bytes to copy
    // from the buffer.
    begin
      Source.Read(b1,1);
      Source.Read(b2,1);

      // Length to copy from buffer.
      Len := (b2 and $f) + 3;
      // Calculate offset in buffer.
      if (b1 >= $f0) and (b2 >= $f0) then
        Idx := (b1 and $f) + 2
      else
        Idx := ((b2 and $f0) shl 4) + b1 + 18;

      // Copy
      for x := 0 to Len-1 do
      begin
        // Get byte from buffer, appent to output & buffer
        c1 := Buffer[Idx and BUFFER_HIGH];
        Buffer[BufferPos and BUFFER_HIGH] := c1;
        Dest.Write(c1,1);

        // Advance local and overall buffer offsets
        Idx := (Idx + 1) and BUFFER_HIGH;
        BufferPos := (BufferPos + 1) and BUFFER_HIGH;
      end;
      Inc(BytesWritten, Len);
    end;

    FlagMask := FlagMask shl 1;
    // See if we used up all 8 bits of the current flag.
    if FlagMask > 255 then
    begin
      Source.Read(Flag, 1);
      FlagMask := 1;
    end;
  end;

  // Rewind desination stream
  Dest.Seek(BytesWritten, soFromEnd);
  Result := BytesWritten;

  if UncompSize <> BytesWritten then
    raise Exception.CreateFmt('Error: Decompressed data is %d bytes, should be %dbytes', [BytesWritten, UncompSize]);
end;


procedure TfrmMain.btnDecompressClick(Sender: TObject);
var
  fin, fout: TFileStream;
begin
  fin := TFileStream.Create(edtSource.Text, fmOpenRead + fmShareDenyNone);
  try
    fout := TFileStream.Create(edtDest.Text, fmCreate or fmShareDenyWrite);
    try
      Decompress(fin, fout);
      ShowMessage('Done..');
    finally
      fout.Free;
    end;
  finally
    fin.Free;
  end;
end;

procedure TfrmMain.btnBrowseForSourceClick(Sender: TObject);
begin
  if OpenDialog.Execute then
    edtSource.Text := OpenDialog.FileName;
end;

procedure TfrmMain.btnBrowseForDestClick(Sender: TObject);
begin
  if SaveDialog.Execute then
    edtDest.Text := SaveDialog.FileName;
end;

procedure TfrmMain.lblHomepageClick(Sender: TObject);
begin
  ShellExecute(handle, 'OPEN', 'http://www.game-editing.net', nil, nil, SW_SHOWNORMAL);
end;

procedure TfrmMain.About1Click(Sender: TObject);
begin
  ShowMessage('By Dan Strandberg && Ryan Foss'+#13#10#13#10+
              'Decompresses the new archive format for Colin McRae Rally 2005.'+#13#10+
              'Inside these arechives are complete old-style archives that'+#13#10+
              'existing tools can open. For example iRipper from Toca Edit'+#13#10+
              '(http://toca.game-editing.net)'+#13#10+#13#10+
              'Thanks to David Skotnicki');
end;

procedure TfrmMain.Exit1Click(Sender: TObject);
begin
  Close;
end;

end.
 